from typing import List
import asyncio
from osgeo import ogr
from .area_types import get_area_types
from .buildings import get_buildings
from .roads import get_roads
from ...utils.helpers.geometry import create_buffered_geometry
from ...utils.constants import DEFAULT_EPSG
from ...models.fact_part import FactPart
from ...models.fact_sheet import FactSheet


async def create_fact_sheet(geometry: ogr.Geometry, orig_epsg: int, buffer: int) -> FactSheet:
    fact_list = await __create_fact_list(geometry, orig_epsg, buffer)

    return FactSheet(None, None, fact_list)


async def __create_fact_list(geometry: ogr.Geometry, orig_epsg: int, buffer: int) -> List[FactPart]:
    input_geom = create_buffered_geometry(geometry, buffer, DEFAULT_EPSG)
    tasks: List[asyncio.Task]

    async with asyncio.TaskGroup() as tg:
        tasks = [
            tg.create_task(get_area_types(
                input_geom, DEFAULT_EPSG, orig_epsg, buffer)),
            tg.create_task(get_buildings(
                input_geom, DEFAULT_EPSG, orig_epsg, buffer)),
            tg.create_task(get_roads(
                input_geom, DEFAULT_EPSG, orig_epsg, buffer))
        ]

    result_list: List[FactPart] = []

    for task in tasks:
        result = task.result()

        if result is not None:
            result_list.append(result)

    return result_list



import socketio
import logging
from .constants import SOCKET_IO_SRV_URL

_LOGGER = logging.getLogger(__name__)

def get_client():
    if not SOCKET_IO_SRV_URL:
        return None
        
    try:
        sio = socketio.SimpleClient()
        sio.connect(SOCKET_IO_SRV_URL, socketio_path='/ws/socket.io')

        return sio
    except Exception as error:
        _LOGGER.warning(error)
        return None



from typing import List
from .fact_part import FactPart

class FactSheet:
    def __init__(self, raster_result: str, cartography: str, fact_list: List[FactPart]):
        self.raster_result = raster_result
        self.cartography = cartography
        self.fact_list = fact_list


import json
from osgeo import ogr, osr
from math import pi
from re import search
from shapely import wkt
from shapely.wkt import dumps
from ..constants import DEFAULT_EPSG, WGS84_EPSG

EARTH_RADIUS = 6371008.8


def geometry_from_gml(gml_str: str) -> ogr.Geometry:
    try:
        return ogr.CreateGeometryFromGML(gml_str)
    except:
        return None


def geometry_from_json(json_str: str) -> ogr.Geometry:
    try:
        return ogr.CreateGeometryFromJson(json_str)
    except:
        return None


def geometry_to_wkt(geometry: ogr.Geometry, epsg: int) -> str:
    wkt_str = geometry.ExportToWkt()
    geometry = wkt.loads(wkt_str)
    coord_precision = 6 if epsg == WGS84_EPSG else 2

    return dumps(geometry, trim=True, rounding_precision=coord_precision)


def geometry_to_arcgis_geom(geometry: ogr.Geometry, epsg: int) -> str:
    if geometry.GetGeometryType() == ogr.wkbMultiPolygon:
        out_geom = ogr.ForceToPolygon(geometry)
    else:
        out_geom = geometry

    coord_precision = 6 if epsg == WGS84_EPSG else 2
    geojson = out_geom.ExportToJson(
        [f'COORDINATE_PRECISION={coord_precision}'])
    obj = json.loads(geojson)

    arcgis_geom = {
        'rings': obj['coordinates'],
        'spatialReference': {
            'wkid': epsg
        }
    }

    return json.dumps(arcgis_geom)


def create_input_geometry(geo_json: dict) -> ogr.Geometry:
    epsg = get_epsg(geo_json)
    geometry = ogr.CreateGeometryFromJson(str(geo_json))

    if epsg != DEFAULT_EPSG:
        return transform_geometry(geometry, epsg, DEFAULT_EPSG)

    return geometry


def create_buffered_geometry(geometry: ogr.Geometry, distance: int, epsg: int) -> ogr.Geometry:
    computed_buffer = length_to_degrees(
        distance) if epsg is None or epsg == WGS84_EPSG else distance

    return geometry.Buffer(computed_buffer, 10)


def transform_geometry(geometry: ogr.Geometry, src_epsg: int, dest_epsg: int) -> ogr.Geometry:
    source = osr.SpatialReference()
    source.ImportFromEPSG(src_epsg)
    source.SetAxisMappingStrategy(osr.OAMS_TRADITIONAL_GIS_ORDER)

    target = osr.SpatialReference()
    target.ImportFromEPSG(dest_epsg)

    transform = osr.CoordinateTransformation(source, target)
    clone: ogr.Geometry = geometry.Clone()
    clone.Transform(transform)

    return clone


def length_to_degrees(distance: float) -> float:
    radians = distance / EARTH_RADIUS
    degrees = radians % (2 * pi)

    return degrees * 180 / pi


def create_run_on_input_geometry_json(geometry: ogr.Geometry, epsg: int, orig_epsg: int) -> dict:
    geom = geometry

    if epsg != orig_epsg:
        geom = transform_geometry(geometry, epsg, orig_epsg)

    coord_precision = 6 if orig_epsg == WGS84_EPSG else 2
    geo_json = json.loads(geom.ExportToJson(
        [f'COORDINATE_PRECISION={coord_precision}']))
    add_geojson_crs(geo_json, epsg)

    return geo_json


def get_epsg(geo_json: dict) -> int:
    crs = geo_json.get('crs', {}).get('properties', {}).get('name')

    if crs is None:
        return WGS84_EPSG

    regex = r'^(http:\/\/www\.opengis\.net\/def\/crs\/EPSG\/0\/|^urn:ogc:def:crs:EPSG::|^EPSG:)(?P<epsg>\d+)$'
    matches = search(regex, crs)

    if matches:
        return int(matches.group('epsg'))

    return WGS84_EPSG


def add_geojson_crs(geojson: dict, epsg: int) -> None:
    if epsg is None or epsg == WGS84_EPSG:
        return

    geojson['crs'] = {
        'type': 'name',
        'properties': {
            'name': 'urn:ogc:def:crs:EPSG::' + str(epsg)
        }
    }
